////////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2001-2025 Alexey Kuryakin daqgroup@mail.ru under MIT license //
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// This file is part of the CRW-DAQ project by DaqGroup - component CRWLIB.   //
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Purpose:                                                                   //
// Engine agent for Lazarus SQLDB.                                            //
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// History:                                                                   //
// 20250609 - Created by A.K.                                                 //
////////////////////////////////////////////////////////////////////////////////

unit _crw_dblaz; //  DB Lazarus - interface for Lazarus SQLDB.

{$I _crw_sysdef.inc}

{$I _crw_sysmode.inc}

{$WARN 5023 off : Unit "$1" not used in $2}

interface

uses
 //////////////////////////////////////////////////////
 {$I _crw_uses_first.inc} // NB: MUST BE FIRST USES !!!
 //////////////////////////////////////////////////////
 sysutils, classes, math, db,  sqldb, bufdataset,
 mssqlconn, pqconnection, oracleconnection, odbcconn,
 mysql40conn, mysql41conn, mysql50conn, mysql51conn,
 mysql55conn, mysql56conn, mysql57conn, mysql80conn,
 sqlite3conn, ibconnection, lclversion,
 xmldatapacketreader,
 Forms, Controls, Graphics, Dialogs,
 _crw_alloc, _crw_ef, _crw_str, _crw_fio, _crw_dynar,
 _crw_dbglog, _crw_proc, _crw_dbcon;

const /////////////////// SQLDB engine identifiers
 db_engine_code_sqldb   = _crw_dbcon.db_engine_code_sqldb; // SQLDB engine code
 db_engine_name_sqldb   = _crw_dbcon.db_engine_name_sqldb; // SQLDB engine name

const /////////////////// List of SQLDB supported databases
 KnownSqlDbProviders = // Providers, i.e. database types:
 { 1  }  'MSSQL'+EOL   // Microsoft SQL
 { 2  } +'Sybase'+EOL  // Subase
 { 3  } +'PQ'+EOL      // PostgreSQL
 { 4  } +'Oracle'+EOL  // Oracle
 { 5  } +'ODBC'+EOL    // ODBC
 { 6  } +'MySQL40'+EOL // MySQL 4.0
 { 7  } +'MySQL41'+EOL // MySQL 4.1
 { 8  } +'MySQL50'+EOL // MySQL 5.0
 { 9  } +'MySQL51'+EOL // MySQL 5.1
 { 10 } +'MySQL55'+EOL // MySQL 5.5
 { 11 } +'MySQL56'+EOL // MySQL 5.6
 { 12 } +'MySQL57'+EOL // MySQL 5.7
 { 13 } +'MySQL80'+EOL // MySQL 8.0
 { 14 } +'SQLite3'+EOL // SQLite3
 { 16 } +'IB'+EOL      // Interbase, Firebird
   ;                   //////////////////////

type
 TSqlStatementTypes = set of TStatementType;

 ///////////////////////////////////////////////////////////////////////////////
 // TSqlDbConnecter - SQLDB connector, include SQL Connection,Transaction,Query.
 ///////////////////////////////////////////////////////////////////////////////
type
 TSqlDbApi = class;
 TSqlDbRecorder = class;
 TSqlDbCommander = class;
 TSqlDbConnecter = class;
 ESqlDbApi = class(Exception);
 TSqlDbMasterObject = class(TMasterObject)
 private
  myParent:TObject;
  function GetParent:TObject;
  procedure SetParent(aParent:TObject);
 public
  property Parent:TObject read GetParent write SetParent;
 protected
  procedure BugReport(E:Exception; O:TObject; const S:LongString);
 end;
 TSqlDbApi = class(TSqlDbMasterObject)
 private
  myConnection  : TSqlConnection;
  myDataSource  : TDataSource;
  myQuery       : TSQLQuery;
  myTrans       : TSQLTransaction;
  myConStrng    : LongString;
  myProvider    : LongString;
  myErrorsList  : TText;
  myErrorsCount : Integer;
  myIsLinked    : Boolean;
 private
  function  GetVerbose:Boolean;
  procedure SetVerbose(aVerbose:Boolean);
  function  GetErrorsList:TText;
  function  GetErrorsCount:Integer;
  procedure SetErrorsCount(aCount:Integer);
  function  GetProvider:LongString;
  function  GetConnection:TSqlConnection;
  function  GetDataSource:TDataSource;
  function  GetQuery:TSQLQuery;
  function  GetTrans:TSQLTransaction;
  function  GetConStrng:LongString;
  procedure SetConStrng(arg:LongString);
  function  GetIsLinked:Boolean;
  function  LinkSqlObjects:Boolean;
  procedure UnlinkSqlObjects;
  procedure FreeSqlObjects;
 public
  constructor Create;
  destructor  Destroy; override;
  procedure   AfterConstruction; override;
  procedure   BeforeDestruction; override;
 public
  property Verbose:Boolean           read GetVerbose write SetVerbose;
  property ErrorsList:TText          read GetErrorsList;
  property ErrorsCount:Integer       read GetErrorsCount write SetErrorsCount;
  property IsLinked:Boolean          read GetIsLinked;
  property Provider:LongString       read GetProvider;
  property Connection:TSqlConnection read GetConnection;
  property DataSource:TDataSource    read GetDataSource;
  property Query:TSQLQuery           read GetQuery;
  property Trans:TSQLTransaction     read GetTrans;
  property ConStrng:LongString       read GetConStrng write SetConStrng;
 public
  function IncErrorsCount:Integer;
  function MakeConnection(const aConnStr:LongString):Boolean;
 public // Parse parameters
  class function FetchParam(const Buff,Items:LongString):LongString;
  class function ValidateParams(const aBuff,aIncl,aExcl:LongString;
                                const aDelim:LongString=EOL):LongString;
 public // String identifier alternatives
  class function sia_Driver:LongString;
  class function sia_Provider:LongString;
  class function sia_DataBase:LongString;
  class function sia_HostName:LongString;
  class function sia_UserName:LongString;
  class function sia_Password:LongString;
  class function sia_Charset:LongString;
  class function sia_Role:LongString;
  class function sia_Verbose:LongString;
  class function sia_Internals:LongString;
 public // Helper functions
  class procedure ErrorNilReference(const Where:LongString);
  class procedure ErrorNotSupported(const Where:LongString);
  class procedure ParameterUnused(x:Boolean); overload;
  class procedure ParameterUnused(x:Integer); overload;
  class procedure ParameterUnused(x:LongString); overload;
 public
  class var DefVerbose:Boolean;
  class var RaiseOnNilReference:Boolean;
  class var RaiseOnResetProvider:Boolean;
  class var RaiseOnResetConStrng:Boolean;
 end;
 //////////////////
 // TSqlDbConnecter
 //////////////////
 TSqlDbConnecter = class(TSqlDbMasterObject)
 private
  myApi       : TSqlDbApi;
  myCommander : TSqlDbCommander;
  myRecorder  : TSqlDbRecorder;
 private
  function  GetApi:TSqlDbApi;
  function  GetProvider:LongString;
  procedure SetProvider(p:LongString);
  function  GetRecorder:TSqlDbRecorder;
  function  GetCommander:TSqlDbCommander;
 public
  constructor Create(aApi:TSqlDbApi);
  destructor  Destroy; override;
  procedure   AfterConstruction; override;
  procedure   BeforeDestruction; override;
 public
  property Api:TSqlDbApi read GetApi;
  property Provider:LongString read GetProvider write SetProvider;
  property Recorder:TSqlDbRecorder read GetRecorder;
  property Commander:TSqlDbCommander read GetCommander;
 public
  function Open(Opt:Integer=0):Boolean;
  function Close:Boolean;
  function Cancel:Boolean;
  function BeginTrans:Integer;
  function CommitTrans(aRetaining:Boolean=False):Boolean;
  function RollbackTrans(aRetaining:Boolean=False):Boolean;
  function Execute(arg:LongString; opt:Integer=0; rof:Boolean=True):TSqlDbRecorder;
  function State:Integer;
  function Active:Boolean;
 private
  function  GetQuerySQLText:LongString;
  procedure SetQuerySQLText(const arg:LongString);
 public
  property  QuerySQLText:LongString read GetQuerySQLText write SetQuerySQLText;
 public
  procedure OnDbLog(Sender:TSQLConnection; EventType:TDBEventType; const Msg:String);
 public
  class var DefLogEvents:TDBEventTypes;
 private
  myPropertiesWanted:LongString;
 private // Properties
  function  GetVersion:LongString;
  function  GetDefaultDataBase:LongString;
  procedure SetDefaultDataBase(aDatabase:LongString);
  function  GetProperties:LongString;
  procedure SetProperties(p:LongString);
 private // ADO simulation internals
  function  GetConnectionTimeout:Integer;
  procedure SetConnectionTimeout(t:Integer);
  function  GetCommandTimeout:Integer;
  procedure SetCommandTimeout(t:Integer);
  function  GetIsolationLevel:Integer;
  procedure SetIsolationLevel(t:Integer);
  function  GetMode:Integer;
  procedure SetMode(m:Integer);
  function  GetCursorLocation:Integer;
  procedure SetCursorLocation(p:Integer);
  function  GetAttributes:Integer;
  procedure SetAttributes(a:Integer);
 public  // ADO simulation properties
  property  ConnectionTimeout:Integer read GetConnectionTimeout write SetConnectionTimeout;
  property  CommandTimeout:Integer read GetCommandTimeout write SetCommandTimeout;
  property  IsolationLevel:Integer read GetIsolationLevel write SetIsolationLevel;
  property  DefaultDataBase:LongString read GetDefaultDataBase write SetDefaultDataBase;
  property  Version:LongString read GetVersion;
  property  Properties:LongString read GetProperties write SetProperties;
  property  Mode:Integer read GetMode write SetMode;
  property  CursorLocation:Integer read GetCursorLocation write SetCursorLocation;
  property  Attributes:Integer read GetAttributes write SetAttributes;
 public
  function  GetErrors(aClear:Boolean=False):LongString;
  function  GetErrorsCount(aClear:Boolean=False):Integer;
 public // Helper functions
  class function FieldTypeToCode(T:TFieldType):Integer;
 public
  class var DefPropertiesWanted:LongString;
  class var SelectableStatementTypes:TSqlStatementTypes;
 end;
 //////////////////
 // TSqlDbCommander
 //////////////////
 TSqlDbCommander = class(TSqlDbMasterObject)
 private
  myConnecter:TSqlDbConnecter;
 private
  function  GetConnecter:TSqlDbConnecter;
 public
  constructor Create(aConnecter:TSqlDbConnecter);
 public
  property Connecter:TSqlDbConnecter read GetConnecter;
 public
  function  State:Integer;
  function  Active:Boolean;
  function  Cancel:Boolean;
 private
  function  GetCommandType:Integer;
  procedure SetCommandType(t:Integer);
  function  GetCommandText:LongString;
  procedure SetCommandText(aCommand:LongString);
 public
  property  CommandType:Integer read GetCommandType write SetCommandType;
  property  CommandText:LongString read GetCommandText write SetCommandText;
 end;
 /////////////////
 // TSqlDbRecorder
 /////////////////
 TSqlDbRecorder = class(TSqlDbMasterObject)
 private
  myConnecter:TSqlDbConnecter;
 private
  function  GetConnecter:TSqlDbConnecter;
  function  GetFieldsCount:Integer;
  function  GetFieldsNames(i:Integer):LongString;
  function  GetFieldsTypes(i:Integer):Integer;
  function  GetFieldsAsString(i:Integer):LongString;
  procedure SetFieldsAsString(i:Integer; const v:LongString);
  function  GetFieldsAsInteger(i:Integer):Integer;
  procedure SetFieldsAsInteger(i:Integer; v:Integer);
  function  GetFieldsAsFloat(i:Integer):Double;
  procedure SetFieldsAsFloat(i:Integer; v:Double);
 public
  constructor Create(aConnecter:TSqlDbConnecter);
 public
  property  Connecter:TSqlDbConnecter read GetConnecter;
 public // DB API functions
  function  BOF:Boolean;
  function  EOF:Boolean;
  function  MoveFirst:Boolean;
  function  MoveLast:Boolean;
  function  MoveNext:Boolean;
  function  MovePrevious:Boolean;
  function  Close:Boolean;
  function  Open(opt:Integer):Boolean;
  function  Update:Boolean;
  function  Cancel:Boolean;
  function  CancelUpdate:Boolean;
  function  State:Integer;
  function  Active:Boolean;
  function  Requery(opt:Integer=0):Boolean;
  function  Resync(aff,res:Integer):Boolean;
  function  Delete(aff:Integer):Boolean;
  function  AddNew(arg:LongString):Boolean;
  function  IndexOfField(const id:LongString):Integer;
  function  GetString(n:Integer; coldel,rowdel,nullexpr:LongString):LongString;
  function  Supports(CursorOptions:Integer):Boolean;
 public // Data access properties
  property  FieldsCount:Integer read GetFieldsCount;
  property  FieldsNames[i:Integer]:LongString read GetFieldsNames;
  property  FieldsTypes[i:Integer]:Integer read GetFieldsTypes;
  property  FieldsAsString[i:Integer]:LongString read GetFieldsAsString write SetFieldsAsString;
  property  FieldsAsInteger[i:Integer]:Integer read GetFieldsAsInteger write SetFieldsAsInteger;
  property  FieldsAsFloat[i:Integer]:Double read GetFieldsAsFloat write SetFieldsAsFloat;
 private
  function  GetSource:LongString;
  procedure SetSource(aSource:LongString);
 public
  property  Source:LongString read GetSource write SetSource;
 private
  function  GetAbsolutePage:Integer;
  procedure SetAbsolutePage(p:Integer);
  function  GetAbsolutePosition:Integer;
  procedure SetAbsolutePosition(p:Integer);
  function  GetBookmark:Integer;
  procedure SetBookmark(p:Integer);
  function  GetCacheSize:Integer;
  procedure SetCacheSize(p:Integer);
  function  GetCursorType:Integer;
  procedure SetCursorType(p:Integer);
  function  GetEditMode:Integer;
  procedure SetEditMode(p:Integer);
  function  GetFilter:LongString;
  procedure SetFilter(v:LongString);
  function  GetIndex:LongString;
  procedure SetIndex(v:LongString);
  function  GetLockType:Integer;
  procedure SetLockType(p:Integer);
  function  GetMarshalOptions:Integer;
  procedure SetMarshalOptions(p:Integer);
  function  GetMaxRecords:Integer;
  procedure SetMaxRecords(p:Integer);
  function  GetPageCount:Integer;
  function  GetPageSize:Integer;
  procedure SetPageSize(p:Integer);
  function  GetRecordCount:Integer;
  function  GetSort:LongString;
  procedure SetSort(v:LongString);
  function  GetStatus:Integer;
  function  GetStayInSync:Boolean;
  procedure SetStayInSync(v:Boolean);
  function  GetCursorLocation:Integer;
  procedure SetCursorLocation(p:Integer);
 public
  property  AbsolutePage:Integer read GetAbsolutePage write SetAbsolutePage;
  property  AbsolutePosition:Integer read GetAbsolutePosition write SetAbsolutePosition;
  property  Bookmark:Integer read GetBookmark write SetBookmark;
  property  CacheSize:Integer read GetCacheSize write SetCacheSize;
  property  CursorType:Integer read GetCursorType write SetCursorType;
  property  EditMode:Integer read GetEditMode write SetEditMode;
  property  Filter:LongString read GetFilter write SetFilter;
  property  Index:LongString read GetIndex write SetIndex;
  property  LockType:Integer read GetLockType write SetLockType;
  property  MarshalOptions:Integer read GetMarshalOptions write SetMarshalOptions;
  property  MaxRecords:Integer read GetMaxRecords write SetMaxRecords;
  property  PageCount:Integer read GetPageCount;
  property  PageSize:Integer read GetPageSize write SetPageSize;
  property  RecordCount:Integer read GetRecordCount;
  property  Sort:LongString read GetSort write SetSort;
  property  Status:Integer read GetStatus;
  property  StayInSync:Boolean read GetStayInSync write SetStayInSync;
  property  CursorLocation:Integer read GetCursorLocation write SetCursorLocation;
 public
  function  Save(Destination:LongString; PersistFormat:Integer):Boolean;
  function  SaveAsText(Dest,coldel,rowdel,nullexpr:LongString;Quote:Char):Boolean;
 end;

function  NewSqlDbConnecter(const aConnStr:LongString):TSqlDbConnecter;
procedure Kill(var TheObject:TSqlDbConnecter); overload;
procedure Kill(var TheObject:TSqlDbCommander); overload;
procedure Kill(var TheObject:TSqlDbRecorder); overload;
procedure Kill(var TheObject:TSqlDbApi); overload;

function AllSqlDbConnecters:TObjectStorage;

 ////////////////////////////////////////////////////////
 // Get list of ODBC drivers by command 'odbcinst -q -d'.
 // This feature depends on deb package 'odbcinst'.
 ////////////////////////////////////////////////////////
function ReadUnixOdbcDriverList(Update:Boolean=False):LongString;

 /////////////////////////////////////////////////////
 // Get list of ODBC drivers by command 'odbcinst -j'.
 // This feature depends on deb package 'odbcinst'.
 // Return Name=Value list with names from list of
 // UnixOdbcConfigParamsNames.
 /////////////////////////////////////////////////////
function ReadUnixOdbcConfigParams(Update:Boolean=False):LongString;

 /////////////////////////////////////////////////////
 // List of Names of UnixODBC configuration parameters
 /////////////////////////////////////////////////////
const
 UnixOdbcConfigParamsNames='DRIVERS'+EOL
                          +'SYSTEM DATA SOURCES'+EOL
                          +'FILE DATA SOURCES'+EOL
                          +'USER DATA SOURCES'+EOL
                          +'SQLULEN Size'+EOL
                          +'SQLLEN Size'+EOL
                          +'SQLSETPOSIROW Size'+EOL;

 ////////////////////
 // DebugLog channels
 ////////////////////
function dlc_SqlDbBug:Integer;
function dlc_SqlDbLog:Integer;

implementation

uses
 _crw_dbapi;

//////////////////
// Common routines
//////////////////

function dlc_SqlDbBug:Integer;
const dlc:Integer=0;
begin
 if (dlc=0) then dlc:=RegisterDebugLogChannel('_SqlDbBug');
 Result:=dlc;
end;

function dlc_SqlDbLog:Integer;
const dlc:Integer=0;
begin
 if (dlc=0) then dlc:=RegisterDebugLogChannel('_SqlDbLog');
 Result:=dlc;
end;

function FormatExceptionMsg(E:Exception; O:TObject; const S:LongString):LongString;
var exc,who,msg:LongString;
begin
 if Assigned(E) then exc:=E.ClassName else exc:='ENil';
 if Assigned(O) then who:=O.ClassName else who:='Unknown';
 Result:='Exception '+DoubleAngleQuotedStr(exc)+' from '+DoubleAngleQuotedStr(who);
 if (S<>'') then begin
  msg:=S;
  if StartsText('@!',S) then msg:=SkipWords(1,S,ScanSpaces);
  if (msg<>'') then Result:=Result+' note '+DoubleAngleQuotedStr(msg);
 end;
 if Assigned(E) and (E.Message<>'')
 then Result:=Result+' hint '+DoubleAngleQuotedStr(E.Message);
 if PosEol(Result)>0 then Result:=StringReplace(Result,EOL,' ',[rfReplaceAll]);
end;

procedure BugReport(E:Exception; O:TObject; const S:LongString);
var Api:TSqlDbApi; verb:Integer; msg:LongString;
begin
 TDbConnection.IncTotalBugsCount;
 if (StrFetch(S,1)='@')
 then _crw_alloc.BugReport(E,O,S)
 else _crw_alloc.BugReport(E,O,db_bugreport_prefix+S);
 if Assigned(O) then begin
  if (O is TSqlDbApi) then Api:=TSqlDbApi(O) else
  if (O is TSqlDbConnecter) then Api:=TSqlDbConnecter(O).Api else
  if (O is TSqlDbCommander) then Api:=TSqlDbCommander(O).Connecter.Api else
  if (O is TSqlDbRecorder) then Api:=TSqlDbRecorder(O).Connecter.Api else Api:=nil;
  Api.IncErrorsCount;
  verb:=BoolToInt(Api.Verbose)+BoolToInt(DebugLogEnabled(dlc_SqlDbBug))*2;
  if (verb<>0) then begin
   msg:=FormatExceptionMsg(E,O,S);
   if HasFlags(verb,1) then Api.ErrorsList.AddLn(msg);
   if HasFlags(verb,2) then DebugLog(dlc_SqlDbBug,msg);
  end;
 end;
end;

function  NewSqlDbConnecter(const aConnStr:LongString):TSqlDbConnecter;
var Api:TSqlDbApi;
begin
 Result:=nil;
 try
  Api:=TSqlDbApi.Create;
  if not Api.MakeConnection(aConnStr) then Kill(Api);
  if Assigned(Api) then Result:=TSqlDbConnecter.Create(Api);
 except
  on E:Exception do BugReport(E,nil,'NewSqlDbConnecter');
 end;
end;

procedure Kill(var TheObject:TSqlDbConnecter); overload;
begin
 try
  FreeAndNil(TheObject);
 except
  on E:Exception do BugReport(E,nil,'Kill');
 end;
end;

procedure Kill(var TheObject:TSqlDbCommander); overload;
begin
 try
  FreeAndNil(TheObject);
 except
  on E:Exception do BugReport(E,nil,'Kill');
 end;
end;

procedure Kill(var TheObject:TSqlDbRecorder); overload;
begin
 try
  FreeAndNil(TheObject);
 except
  on E:Exception do BugReport(E,nil,'Kill');
 end;
end;

procedure Kill(var TheObject:TSqlDbApi); overload;
begin
 try
  FreeAndNil(TheObject);
 except
  on E:Exception do BugReport(E,nil,'Kill');
 end;
end;

function ReadUnixOdbcDriverList(Update:Boolean=False):LongString;
var ans,line:LongString; i:Integer;
const cmd='odbcinst -q -d';
const drv:LongString='';
begin
 Result:='';
 if IsUnix then
 try
  ans:='';
  if Update then drv:='';
  if (drv<>'') then Exit(drv);
  if RunCommand(cmd,ans) then
  for i:=1 to WordCount(ans,EolnDelims) do begin
   line:=Trim(ExtractWord(i,ans,EolnDelims));
   if StartsText('[',line) and EndsText(']',line)
   then Result:=Result+Copy(line,2,length(line)-2)+EOL;
  end;
  drv:=Result;
 except
  on E:Exception do BugReport(E,nil,'ReadUnixOdbcDriverList');
 end;
end;

function ReadUnixOdbcConfigParams(Update:Boolean=False):LongString;
var ans,line,sn,sv:LongString; i:Integer;
const cmd='odbcinst -j';
const par:LongString='';
begin
 Result:='';
 if IsUnix then
 try
  ans:=''; sn:=''; sv:='';
  if Update then par:='';
  if (par<>'') then Exit(par);
  if RunCommand(cmd,ans) then
  for i:=1 to WordCount(ans,EolnDelims) do begin
   line:=Trim(ExtractWord(i,ans,EolnDelims));
   if ExtractNameValuePair(line,sn,sv,':')>0 then
   if IsNonEmptyStr(sn) and IsNonEmptyStr(sv) then
   Result:=Result+TrimRightChars(sn,ScanSpaces+['.'])+'='+sv+EOL;
  end;
  par:=Result;
 except
  on E:Exception do BugReport(E,nil,'ReadUnixOdbcConfigParameters');
 end;
end;

////////////////////////////////////
// TSqlDbMasterObject implementation
////////////////////////////////////

function TSqlDbMasterObject.GetParent:TObject;
begin
 if Assigned(Self)
 then Result:=myParent
 else Result:=nil;
end;

procedure TSqlDbMasterObject.SetParent(aParent:TObject);
begin
 if Assigned(Self) and (aParent is TDbConnection)
 then myParent:=aParent;
end;

procedure TSqlDbMasterObject.BugReport(E:Exception; O:TObject; const S:LongString);
var Root:TDbConnection; BgPref:LongString;
begin
 Root:=nil; BgPref:='';
 if Assigned(Self) then begin
  if (myParent is TDbConnection)
  then Root:=TDbConnection(myParent);
  if Assigned(Root) then begin
   BgPref:=Root.Control('BugReportPrefix');
   Root.IncBugsCount;
  end;
 end;
 if (Self=nil)
 then _crw_dblaz.BugReport(E,O,S)
 else _crw_dblaz.BugReport(E,O,BgPref+S);
end;

///////////////////////////
// TSqlDbApi implementation
///////////////////////////

constructor TSqlDbApi.Create;
begin
 inherited Create;
 Verbose:=DefVerbose;
end;

destructor TSqlDbApi.Destroy;
begin
 myConStrng:='';
 myProvider:='';
 FreeSqlObjects;
 Kill(myErrorsList);
 inherited Destroy;
end;

procedure TSqlDbApi.AfterConstruction;
begin
 inherited AfterConstruction;
 LinkSqlObjects;
end;

procedure TSqlDbApi.BeforeDestruction;
begin
 Verbose:=False;
 UnlinkSqlObjects;
 inherited BeforeDestruction;
end;

function TSqlDbApi.GetVerbose:Boolean;
begin
 if Assigned(Self)
 then Result:=Assigned(myErrorsList)
 else Result:=False;
end;

procedure TSqlDbApi.SetVerbose(aVerbose:Boolean);
begin
 if Assigned(Self) then begin
  if not aVerbose then Kill(myErrorsList) else
  if not Assigned(myErrorsList) then begin
   myErrorsList:=NewText;
   myErrorsList.Master:=@myErrorsList;
  end;
 end;
end;

function TSqlDbApi.GetErrorsList:TText;
begin
 if Assigned(Self)
 then Result:=myErrorsList
 else Result:=nil;
end;

function TSqlDbApi.IncErrorsCount:Integer;
begin
 if Assigned(Self)
 then Result:=LockedInc(myErrorsCount)
 else Result:=0;
end;

function TSqlDbApi.GetErrorsCount:Integer;
begin
 if Assigned(Self)
 then Result:=LockedGet(myErrorsCount)
 else Result:=0;
end;

procedure TSqlDbApi.SetErrorsCount(aCount:Integer);
begin
 if Assigned(Self)
 then LockedSet(myErrorsCount,aCount)
end;

function TSqlDbApi.GetProvider:LongString;
begin
 if Assigned(Self)
 then Result:=myProvider
 else Result:='';
end;

function TSqlDbApi.GetConnection:TSqlConnection;
begin
 Result:=nil;
 if Assigned(Self)
 then Result:=myConnection
 else ErrorNilReference('GetConnection');
end;

function TSqlDbApi.GetDataSource:TDataSource;
begin
 Result:=nil;
 if Assigned(Self)
 then Result:=myDataSource
 else ErrorNilReference('GetDataSource');
end;

function TSqlDbApi.GetQuery:TSQLQuery;
begin
 Result:=nil;
 if Assigned(Self)
 then Result:=myQuery
 else ErrorNilReference('GetQuery');
end;

function TSqlDbApi.GetTrans:TSQLTransaction;
begin
 Result:=nil;
 if Assigned(Self)
 then Result:=myTrans
 else ErrorNilReference('GetTrans');
end;

function TSqlDbApi.GetConStrng:LongString;
begin
 if Assigned(Self)
 then Result:=myConStrng
 else Result:='';
end;

procedure TSqlDbApi.SetConStrng(arg:LongString);
begin
 if Assigned(Self) then begin
  if RaiseOnResetConStrng
  then ErrorNotSupported('SetConStrng');
  myConStrng:=arg;
 end;
end;

function TSqlDbApi.GetIsLinked:Boolean;
begin
 if Assigned(Self)
 then Result:=myIsLinked
 else Result:=False;
end;

function TSqlDbApi.LinkSqlObjects:Boolean;
begin
 Result:=False;
 if Assigned(Self) then
 try
  if Assigned(myConnection) then myConnection.Transaction:=myTrans;
  if Assigned(myQuery) then myQuery.Transaction:=myTrans;
  if Assigned(myQuery) then myQuery.SQLConnection:=myConnection;
  if Assigned(myTrans) then myTrans.SQLConnection:=myConnection;
  if Assigned(myDataSource) then myDataSource.DataSet:=myQuery;
  Result:=Assigned(myConnection) and Assigned(myQuery)
      and Assigned(myDataSource) and Assigned(myTrans)
      and (myQuery.SQLConnection=myConnection)
      and (myTrans.SQLConnection=myConnection)
      and (myConnection.Transaction=myTrans)
      and (myDataSource.DataSet=myQuery)
      and (myQuery.Transaction=myTrans);
  myIsLinked:=Result;
 except
  on E:Exception do BugReport(E,Self,'LinkSqlObjects');
 end;
end;

procedure TSqlDbApi.UnlinkSqlObjects;
begin
 if Assigned(Self) then
 try
  myIsLinked:=False;
  if Assigned(myTrans) then myTrans.Active:=False;
  if Assigned(myQuery) then myQuery.Active:=False;
  if Assigned(myConnection) then myConnection.Connected:=False;
  if Assigned(myConnection) then myConnection.Transaction:=nil;
  if Assigned(myQuery) then myQuery.Transaction:=nil;
  if Assigned(myQuery) then myQuery.SQLConnection:=nil;
  if Assigned(myTrans) then myTrans.SQLConnection:=nil;
  if Assigned(myDataSource) then myDataSource.DataSet:=nil;
 except
  on E:Exception do BugReport(E,Self,'UnlinkSqlObjects');
 end;
end;

procedure TSqlDbApi.FreeSqlObjects;
begin
 if Assigned(Self) then
 try
  UnlinkSqlObjects;
  Kill(TObject(myDataSource));
  Kill(TObject(myConnection));
  Kill(TObject(myTrans));
  Kill(TObject(myQuery));
 except
  on E:Exception do BugReport(E,Self,'FreeSqlObjects');
 end;
end;

function TSqlDbApi.MakeConnection(const aConnStr:LongString):Boolean;
var sProvider,sDriver,sDataBase,sHostName,sUserName,sPassword:LongString;
var sCharset,sRole,sVerbose,sConStr,sParams:LongString;
begin
 Result:=false;
 if Assigned(Self) then
 try
  FreeSqlObjects;
  myConStrng:=ValidateParams(aConnStr,'','');
  sConStr:=ValidateParams(aConnStr,sia_Internals,'');
  sParams:=ValidateParams(aConnStr,'',sia_Internals);
  sDriver:=FetchParam(sConStr,sia_Driver);
  sProvider:=FetchParam(sConStr,sia_Provider);
  sDataBase:=FetchParam(sConStr,sia_DataBase);
  sHostName:=FetchParam(sConStr,sia_HostName);
  sUserName:=FetchParam(sConStr,sia_UserName);
  sPassword:=FetchParam(sConStr,sia_Password);
  sCharset:=FetchParam(sConStr,sia_Charset);
  sVerbose:=FetchParam(sConStr,sia_Verbose);
  sRole:=FetchParam(sConStr,sia_Role);
  if IsNonEmptyStr(sProvider) then begin
   // Microsoft SQL
   if (WordIndex(sProvider,'MSSQL',ScanSpaces)>0) then begin
    myConnection:=TMSSQLConnection.Create(nil);
    myProvider:='MSSQL';
   end else
   // Sybase
   if (WordIndex(sProvider,'Sybase',ScanSpaces)>0) then begin
    myConnection:=TSybaseConnection.Create(nil);
    myProvider:='Sybase';
   end else
   // PostgreSQL
   if (WordIndex(sProvider,'PQ PostgreSQL',ScanSpaces)>0) then begin
    myConnection:=TPQConnection.Create(nil);
    myProvider:='PQ';
   end else
   // Oracle
   if (WordIndex(sProvider,'Oracle',ScanSpaces)>0) then begin
    myConnection:=TOracleConnection.Create(nil);
    myProvider:='Oracle';
   end else
   // ODBC
   if (WordIndex(sProvider,'ODBC',ScanSpaces)>0) then begin
    myConnection:=TODBCConnection.Create(nil);
    if (sDriver<>'') then TODBCConnection(myConnection).Driver:=sDriver;
    myProvider:='ODBC';
   end else
   // MySQL 4.0
   if (WordIndex(sProvider,'MySQL40',ScanSpaces)>0) then begin
    myConnection:=TMySQL40Connection.Create(nil);
    myProvider:='MySQL40';
   end else
   // MySQL 4.1
   if (WordIndex(sProvider,'MySQL41',ScanSpaces)>0) then begin
    myConnection:=TMySQL41Connection.Create(nil);
    myProvider:='MySQL41';
   end else
   // MySQL 5.0
   if (WordIndex(sProvider,'MySQL50',ScanSpaces)>0) then begin
    myConnection:=TMySQL50Connection.Create(nil);
    myProvider:='MySQL50';
   end else
   // MySQL 5.1
   if (WordIndex(sProvider,'MySQL51',ScanSpaces)>0) then begin
    myConnection:=TMySQL51Connection.Create(nil);
    myProvider:='MySQL51';
   end else
   // MySQL 5.5
   if (WordIndex(sProvider,'MySQL55',ScanSpaces)>0) then begin
    myConnection:=TMySQL55Connection.Create(nil);
    myProvider:='MySQL55';
   end else
   // MySQL 5.6
   if (WordIndex(sProvider,'MySQL56',ScanSpaces)>0) then begin
    myConnection:=TMySQL56Connection.Create(nil);
    myProvider:='MySQL56';
   end else
   // MySQL 5.7
   if (WordIndex(sProvider,'MySQL57',ScanSpaces)>0) then begin
    myConnection:=TMySQL57Connection.Create(nil);
    myProvider:='MySQL57';
   end else
   // MySQL 8.0
   if (WordIndex(sProvider,'MySQL80',ScanSpaces)>0) then begin
    myConnection:=TMySQL80Connection.Create(nil);
    myProvider:='MySQL80';
   end else
   // SQLite3
   if (WordIndex(sProvider,'SQLite3',ScanSpaces)>0) then begin
    myConnection:=TSQLite3Connection.Create(nil);
    myProvider:='SQLite3';
   end else
   // Interbase, Firebird
   if (WordIndex(sProvider,'IB Interbase Firebird',ScanSpaces)>0) then begin
    myConnection:=TIBConnection.Create(nil);
    myProvider:='IB';
   end;
   if not Assigned(myConnection)
   then Raise ESqlDbApi.CreateFmt('Bad SQLDB Provider (%s).',[sProvider]);
   if (myProvider='') and Assigned(myConnection) then begin
    myProvider:=myConnection.ClassName; System.Delete(myProvider,1,1);
    myProvider:=StringReplace(myProvider,'Connection','',[rfIgnoreCase]);
   end;
   myConnection.LoginPrompt:=false; // For thread safety
   if (sDataBase<>'') then myConnection.DatabaseName:=sDataBase;
   if (sHostName<>'') then myConnection.HostName:=sHostName;
   if (sUserName<>'') then myConnection.UserName:=sUserName;
   if (sPassword<>'') then myConnection.Password:=sPassword;
   if (sCharset<>'') then myConnection.CharSet:=sCharset;
   if (sVerbose<>'') then Verbose:=(StrToIntDef(sVerbose,BoolToInt(Verbose))>0);
   if (sRole<>'') then myConnection.Role:=sRole;
   if (sParams<>'') then myConnection.Params.Text:=sParams;
   if (myDataSource=nil) then myDataSource:=TDataSource.Create(nil);
   if (myTrans=nil) then myTrans:=TSQLTransaction.Create(nil);
   if (myQuery=nil) then myQuery:=TSQLQuery.Create(nil);
   Result:=LinkSqlObjects;
  end;
 except
  on E:Exception do BugReport(E,Self,'MakeConnection');
 end;
end;

class function TSqlDbApi.FetchParam(const Buff,Items:LongString):LongString;
begin
 Result:=Trim(CookieScanAlter(Buff,Items,Ord(';')));
end;

class function TSqlDbApi.ValidateParams(const aBuff,aIncl,aExcl:LongString;
                                        const aDelim:LongString=EOL):LongString;
var Lines:TStringList; sn,sv:LongString; i:Integer;
const Delims=EolnDelims+[';',','];
begin
 Result:='';
 try
  sn:='';sv:='';
  Lines:=TStringList.Create;
  try
   Lines.Text:=StringReplace(aBuff,';',EOL,[rfReplaceAll]);
   for i:=Lines.Count-1 downto 0 do begin
    if (ExtractNameValuePair(Lines[i],sn,sv)>0) then begin
     if (aIncl<>'') then begin
      if (WordIndex(sn,aIncl,Delims)=0) then begin sn:=''; sv:=''; end;
     end;
     if (aExcl<>'') then begin
      if (WordIndex(sn,aExcl,Delims)>0) then begin sn:=''; sv:=''; end;
     end;
    end else begin
     sn:=''; sv:='';
    end;
    if (sn<>'') and (sv<>'')
    then Lines[i]:=Format('%s=%s',[sn,sv])
    else Lines.Delete(i);
   end;
   Result:=Lines.Text;
   if (aDelim<>EOL) then Result:=StringReplace(Result,EOL,aDelim,[rfReplaceAll]);
  finally
   Kill(Lines);
  end;
 except
  on E:Exception do _crw_dblaz.BugReport(E,nil,'ValidateParams');
 end;
end;

class function TSqlDbApi.sia_Driver:LongString;
begin
 Result:='Driver';
end;
class function TSqlDbApi.sia_Provider:LongString;
begin
 Result:='Provider';
end;
class function TSqlDbApi.sia_DataBase:LongString;
begin
 Result:='DataBase;Location;Data Source;DBNAME;FILEDSN;DSN;DataBaseName';
end;
class function TSqlDbApi.sia_HostName:LongString;
begin
 Result:='HostName;DataSource;Server;Host';
end;
class function TSqlDbApi.sia_UserName:LongString;
begin
 Result:='UserName;User Name;User_Name;UserId;User ID;User;UID';
end;
class function TSqlDbApi.sia_Password:LongString;
begin
 Result:='Password;PWD';
end;
class function TSqlDbApi.sia_Charset:LongString;
begin
 Result:='Charset';
end;
class function TSqlDbApi.sia_Role:LongString;
begin
 Result:='Role';
end;
class function TSqlDbApi.sia_Verbose:LongString;
begin
 Result:='Verbose';
end;
class function TSqlDbApi.sia_Internals:LongString;
const Sep=';'; // Separator
begin
 Result:=sia_Driver+Sep+sia_Provider+Sep+sia_DataBase+Sep
        +sia_HostName+Sep+sia_UserName+Sep+sia_Password+Sep
        +sia_Charset+Sep+sia_Role+Sep+sia_Verbose;
end;

class procedure TSqlDbApi.ErrorNilReference(const Where:LongString);
begin
 if RaiseOnNilReference then
 raise ESqlDbApi.Create(Format('Nil reference at %s.',[Where]));
end;

class procedure TSqlDbApi.ErrorNotSupported(const Where:LongString);
begin
 raise ESqlDbApi.Create(Format('SQLDB not supports %s.',[Where]));
end;

class procedure TSqlDbApi.ParameterUnused(x:Boolean);
begin
 FakeNop(x);
end;

class procedure TSqlDbApi.ParameterUnused(x:Integer);
begin
 FakeNop(x);
end;

class procedure TSqlDbApi.ParameterUnused(x:LongString);
begin
 FakeNop(x);
end;

/////////////////////////////////
// TSqlDbConnecter implementation
/////////////////////////////////

constructor TSqlDbConnecter.Create(aApi:TSqlDbApi);
begin
 inherited Create;
 if Assigned(aApi) then begin
  myApi:=aApi;
  myApi.Master:=@myApi;
 end;
 myCommander:=TSqlDbCommander.Create(Self);
 myCommander.Master:=@myCommander;
 myRecorder:=TSqlDbRecorder.Create(Self);
 myRecorder.Master:=@myRecorder;
 myPropertiesWanted:=DefPropertiesWanted;
end;

destructor TSqlDbConnecter.Destroy;
begin
 myPropertiesWanted:='';
 Kill(myCommander);
 Kill(myRecorder);
 Kill(myApi);
 inherited Destroy;
end;

procedure TSqlDbConnecter.AfterConstruction;
begin
 inherited AfterConstruction;
 AllSqlDbConnecters.Add(Self);
 if Api.IsLinked then begin
  if DebugLogEnabled(dlc_SqlDbLog)
  then Api.Connection.OnLog:=OnDbLog;
  Api.Connection.LogEvents:=DefLogEvents;
 end;
end;

procedure TSqlDbConnecter.BeforeDestruction;
begin
 AllSqlDbConnecters.Remove(Self);
 inherited BeforeDestruction;
end;

function TSqlDbConnecter.GetApi:TSqlDbApi;
begin
 if Assigned(Self)
 then Result:=myApi
 else Result:=nil;
end;

function TSqlDbConnecter.GetRecorder:TSqlDbRecorder;
begin
 if Assigned(Self)
 then Result:=myRecorder
 else Result:=nil;
end;

function TSqlDbConnecter.GetCommander:TSqlDbCommander;
begin
 if Assigned(Self)
 then Result:=myCommander
 else Result:=nil;
end;

function TSqlDbConnecter.GetProvider:LongString;
begin
 if Assigned(Self)
 then Result:=myApi.Provider
 else Result:='';
end;

procedure TSqlDbConnecter.SetProvider(p:LongString);
begin
 if Assigned(Self) then
 try
  if Api.RaiseOnResetProvider
  then Api.ErrorNotSupported('SetProvider');
  Api.ParameterUnused(p);
 except
  on E:Exception do BugReport(E,Self,'SetProvider');
 end;
end;

function TSqlDbConnecter.Open(Opt:Integer=0):Boolean;
begin
 Result:=False;
 if Assigned(Self) then
 try
  if Api.IsLinked then begin
   if not Api.Connection.Connected
   then Api.Connection.Connected:=True;
   Result:=Api.Connection.Connected;
  end;
  Api.ParameterUnused(Opt);
 except
  on E:Exception do BugReport(E,Self,'Open');
 end;
end;

function TSqlDbConnecter.Close:Boolean;
begin
 Result:=False;
 if Assigned(Self) then
 try
  if Api.IsLinked then begin
   if Api.Query.Active then Api.Query.Active:=False;
   if Api.Trans.Active then Api.Trans.Active:=False;
   if Api.Connection.Connected then Api.Connection.Connected:=False;
   Result:=not Api.Connection.Connected;
  end;
 except
  on E:Exception do BugReport(E,Self,'Close');
 end;
end;

function TSqlDbConnecter.Cancel:Boolean;
begin
 Result:=False;
 if Assigned(Self) then
 try
  if Api.IsLinked then begin
   if Api.Query.Active then begin
    Api.DataSource.DataSet.Cancel;
    Api.Query.CancelUpdates;
    Result:=true;
   end;
  end;
 except
  on E:Exception do BugReport(E,Self,'Cancel');
 end;
end;

function TSqlDbConnecter.BeginTrans:Integer;
begin
 Result:=0;
 if Assigned(Self) then
 try
  if Api.IsLinked then
  if Api.Connection.Connected then begin
   if not Api.Connection.Transaction.Active
   then Api.Connection.Transaction.Active:=True;
   Result:=BoolToInt(Api.Connection.Transaction.Active);
  end;
 except
  on E:Exception do BugReport(E,Self,'BeginTrans');
 end;
end;

function TSqlDbConnecter.CommitTrans(aRetaining:Boolean=False):Boolean;
begin
 Result:=False;
 if Assigned(Self) then
 try
  if Api.IsLinked then
  if Api.Connection.Connected then
  if Api.Connection.Transaction.Active then begin
   if not aRetaining
   then Api.Connection.Transaction.Commit
   else Api.Connection.Transaction.CommitRetaining;
   Result:=True;
  end;
 except
  on E:Exception do BugReport(E,Self,'CommitTrans');
 end;
end;

function TSqlDbConnecter.RollbackTrans(aRetaining:Boolean=False):Boolean;
begin
 Result:=False;
 if Assigned(Self) then
 try
  if Api.IsLinked then
  if Api.Connection.Connected then
  if Api.Connection.Transaction.Active then begin
   if not aRetaining
   then Api.Connection.Transaction.Rollback
   else Api.Connection.Transaction.RollbackRetaining;
   Result:=True;
  end;
 except
  on E:Exception do BugReport(E,Self,'RollbackTrans');
 end;
end;

function TSqlDbConnecter.Execute(arg:LongString; opt:Integer=0; rof:Boolean=True):TSqlDbRecorder;
var st:TStatementType;
begin
 Result:=nil;
 if Assigned(Self) then
 try
  if Api.IsLinked then
  if Api.Connection.Connected then begin
   QuerySQLText:=arg;
   st:=Api.Query.StatementType;
   // Query.Open raise error on non-selectable SQL
   // So (re)open on selectable SQL or use ExecSQL
   if (st in SelectableStatementTypes) then begin
    if Api.Query.Active
    then Api.Query.Active:=False;
    Api.Query.ReadOnly:=rof;
    Api.Query.Active:=true;
   end else begin
    Api.Query.ExecSQL;
   end;
   Result:=myRecorder;
  end;
  Api.ParameterUnused(Opt);
 except
  on E:Exception do BugReport(E,Self,'Execute');
 end;
end;

function TSqlDbConnecter.State:Integer;
begin
 Result:=adStateClosed;
 if Assigned(Self) then
 try
  if Api.IsLinked then
  if Api.Connection.Connected then begin
   Result:=adStateOpen;
  end;
 except
  on E:Exception do BugReport(E,Self,'State');
 end;
end;

function TSqlDbConnecter.Active:Boolean;
begin
 Result:=False;
 if Assigned(Self) then
 try
  if Api.IsLinked then
  if Api.Connection.Connected then begin
   Result:=True;
  end;
 except
  on E:Exception do BugReport(E,Self,'Active');
 end;
end;

function TSqlDbConnecter.GetQuerySQLText:LongString;
begin
 Result:='';
 if Assigned(Self) then
 try
  if Api.IsLinked then begin
   Result:=Trim(Api.Query.SQL.Text);
  end;
 except
  on E:Exception do BugReport(E,Self,'GetQuerySQLText');
 end;
end;

procedure TSqlDbConnecter.SetQuerySQLText(const arg:LongString);
begin
 if Assigned(Self) then
 try
  if Api.IsLinked then begin
   Api.Query.SQL.Text:=Trim(arg);
   if not Api.Query.Prepared
   then Api.Query.Prepare;
  end;
 except
  on E:Exception do BugReport(E,Self,'SetQuerySQLText');
 end;
end;

procedure TSqlDbConnecter.OnDbLog(Sender:TSQLConnection; EventType:TDBEventType; const Msg:String);
var Evt:LongString;
begin
 if Assigned(Sender) then
 if DebugLogEnabled(dlc_SqlDbLog) then begin
  Evt:=DbCon.EventTypeToString(EventType);
  if StartsText('det',Evt) then Delete(Evt,1,3);
  DebugLog(dlc_SqlDbLog,Format('%s %s - %s',[Sender.ClassName,Evt,Msg]));
 end;
end;

function TSqlDbConnecter.GetDefaultDataBase:LongString;
begin
 Result:='';
 if Assigned(Self) then
 try
  if Api.IsLinked then begin
   Result:=Api.Connection.DatabaseName;
  end;
 except
  on E:Exception do BugReport(E,Self,'GetDefaultDataBase');
 end;
end;

procedure TSqlDbConnecter.SetDefaultDataBase(aDatabase:LongString);
begin
 if Assigned(Self) then
 try
  if Api.IsLinked then begin
   Api.Connection.DatabaseName:=aDatabase;
  end;
 except
  on E:Exception do BugReport(E,Self,'SetDefaultDataBase');
 end;
end;

function TSqlDbConnecter.GetVersion:LongString;
begin
 if Assigned(Self)
 then Result:=db_engine_name_sqldb+' LCL '+lcl_version+' FPC '+GetFpcVersion
 else Result:='';
end;

const
 AllPropertiesAvailable='Connection,Transaction,Query,DataSource,Database,Details,Defaults,Reset,All';

function TSqlDbConnecter.GetProperties:LongString;
var Lines,List:TStringList; Details:Boolean;
 procedure AddProp(aName,aValue:LongString);
 begin
  aName:=Trim(aName);
  aValue:=Trim(aValue);
  if (aName<>'') // and (aValue<>'')
  then Lines.Values[aName]:=aValue;
 end;
begin
 Result:='';
 // Stub for ADO simulation
 if Assigned(Self) then
 try
  if Api.IsLinked then begin
   List:=TStringList.Create;
   Lines:=TStringList.Create;
   try
    List.Delimiter:=',';
    AddProp('EngineId',IntToStr(db_engine_code_sqldb));
    AddProp('EngineName',db_engine_name_sqldb);
    AddProp('EngineVersion',Version);
    AddProp('Properties.Wanted',myPropertiesWanted);
    AddProp('Properties.Defaults',DefPropertiesWanted);
    AddProp('Properties.All',AllPropertiesAvailable);
    Details:=(WordIndex('Details',myPropertiesWanted,ScanSpaces)>0);
    if WordIndex('Connection',myPropertiesWanted,ScanSpaces)>0 then begin
     AddProp('Connection.Connected',IntToStr(BoolToInt(Api.Connection.Connected)));
     AddProp('Connection.DatabaseName',Api.Connection.DatabaseName);
     AddProp('Connection.UserName',Api.Connection.UserName);
     AddProp('Connection.Password',Api.Connection.Password);
     AddProp('Connection.HostName',Api.Connection.HostName);
     AddProp('Connection.CharSet',Api.Connection.CharSet);
     AddProp('Connection.Role',Api.Connection.Role);
     AddProp('Connection.ConnOptions',UnifySection(DbCon.ConnOptionsToString(Api.Connection.ConnOptions),0));
     AddProp('Connection.Options',UnifySection(DbCon.SQLConnectionOptionsToString(Api.Connection.Options),0));
     if Details then AddProp('Connection.ConnectionInfo',Api.Connection.GetConnectionInfo(citAll));
    end;
    if WordIndex('Transaction',myPropertiesWanted,ScanSpaces)>0 then begin
     AddProp('Transaction.Active',BoolToStr(Api.Trans.Active,'1','0'));
    end;
    if WordIndex('Query',myPropertiesWanted,ScanSpaces)>0 then begin
     AddProp('Query.Active',BoolToStr(Api.Query.Active,'1','0'));
     AddProp('Query.StatementType',DbCon.StatementTypeToString(Api.Query.StatementType));
     if Details then AddProp('Query.SQLText',QuerySQLText);
     if Details then AddProp('Query.RowsAffected',IntToStr(Api.Query.RowsAffected));
     if Details then AddProp('Query.SchemaType',DbCon.SchemaTypeToString(Api.Query.SchemaType));
    end;
    if WordIndex('DataSource',myPropertiesWanted,ScanSpaces)>0 then begin
     AddProp('DataSource.State',DbCon.DataSetStateToString(Api.DataSource.State));
    end;
    if WordIndex('Database',myPropertiesWanted,ScanSpaces)>0 then begin
     if Api.Connection.Connected then begin
      Api.Connection.GetTableNames(List);
      AddProp('TableNames',List.CommaText);
      List.Clear;
     end;
    end;
    if WordIndex('Reset',myPropertiesWanted,ScanSpaces)>0 then begin
     myPropertiesWanted:=DefPropertiesWanted;
    end;
    Result:=Lines.Text;
   finally
    Kill(Lines);
    Kill(List);
   end;
  end;
 except
  on E:Exception do BugReport(E,Self,'GetProperties');
 end;
end;

procedure TSqlDbConnecter.SetProperties(p:LongString);
var Lines:TStringList; sv:LongString; iv:Integer;
begin
 // Stub for ADO simulation
 if Assigned(Self) then
 try
  if Api.IsLinked then begin
   Lines:=TStringList.Create;
   try
    Lines.Text:=p;
    sv:=Lines.Values['Properties.Wanted'];
    if (sv<>'') then begin
     sv:=StringReplace(sv,'Defaults',DefPropertiesWanted,[rfReplaceAll,rfIgnoreCase]);
     if (WordIndex('All',sv,ScanSpaces)>0) then sv:=AllPropertiesAvailable;
     myPropertiesWanted:=Trim(sv);
    end;
    sv:=Lines.Values['Connection.DatabaseName'];
    if (sv<>'') then Api.Connection.DatabaseName:=Trim(sv);
    sv:=Lines.Values['Connection.UserName'];
    if (sv<>'') then Api.Connection.UserName:=Trim(sv);
    sv:=Lines.Values['Connection.Password'];
    if (sv<>'') then Api.Connection.Password:=Trim(sv);
    sv:=Lines.Values['Connection.HostName'];
    if (sv<>'') then Api.Connection.HostName:=Trim(sv);
    sv:=Lines.Values['Connection.CharSet'];
    if (sv<>'') then Api.Connection.CharSet:=Trim(sv);
    sv:=Lines.Values['Connection.Role'];
    if (sv<>'') then Api.Connection.Role:=Trim(sv);
    sv:=Lines.Values['Connection.Connected'];
    if (sv<>'') and TryStrToInt(sv,iv) then Api.Connection.Connected:=(iv>0);
   finally
    Kill(Lines);
   end;
  end;
 except
  on E:Exception do BugReport(E,Self,'SetProperties');
 end;
end;

function TSqlDbConnecter.GetConnectionTimeout:Integer;
begin
 Result:=0;
 // Stub for ADO simulation
end;

procedure TSqlDbConnecter.SetConnectionTimeout(t:Integer);
begin
 // Stub for ADO simulation
 if Assigned(Self) then Api.ParameterUnused(t);
end;

function TSqlDbConnecter.GetCommandTimeout:Integer;
begin
 Result:=0;
 // Stub for ADO simulation
end;

procedure TSqlDbConnecter.SetCommandTimeout(t:Integer);
begin
 // Stub for ADO simulation
 if Assigned(Self) then Api.ParameterUnused(t);
end;

function TSqlDbConnecter.GetIsolationLevel:Integer;
begin
 Result:=0;
 // Stub for ADO simulation
end;

procedure TSqlDbConnecter.SetIsolationLevel(t:Integer);
begin
 // Stub for ADO simulation
 if Assigned(Self) then Api.ParameterUnused(t);
end;

function TSqlDbConnecter.GetMode:Integer;
begin
 Result:=0;
 // Stub for ADO simulation
end;

procedure TSqlDbConnecter.SetMode(m:Integer);
begin
 // Stub for ADO simulation
 if Assigned(Self) then Api.ParameterUnused(m);
end;

function TSqlDbConnecter.GetCursorLocation:Integer;
begin
 Result:=0;
 // Stub for ADO simulation
end;

procedure TSqlDbConnecter.SetCursorLocation(p:Integer);
begin
 // Stub for ADO simulation
 if Assigned(Self) then Api.ParameterUnused(p);
end;

function TSqlDbConnecter.GetAttributes:Integer;
begin
 Result:=0;
 // Stub for ADO simulation
end;

procedure TSqlDbConnecter.SetAttributes(a:Integer);
begin
 // Stub for ADO simulation
 if Assigned(Self) then Api.ParameterUnused(a);
end;

function TSqlDbConnecter.GetErrors(aClear:Boolean=False):LongString;
begin
 Result:='';
 if Assigned(Self) then
 if Assigned(Api) then begin
  Result:=Api.ErrorsList.Text;
  if aClear then GetErrorsCount(aClear);
 end;
end;

function TSqlDbConnecter.GetErrorsCount(aClear:Boolean=False):Integer;
begin
 Result:=0;
 if Assigned(Self) then
 if Assigned(Api) then begin
  Result:=Api.ErrorsCount;
  if aClear then Api.ErrorsCount:=0;
  if aClear and Api.Verbose then Api.ErrorsList.Count:=0;
 end;
end;

class function TSqlDbConnecter.FieldTypeToCode(T:TFieldType):Integer;
begin
 Result:=Ord(T);
end;

/////////////////////////////////
// TSqlDbCommander implementation
/////////////////////////////////

constructor TSqlDbCommander.Create(aConnecter:TSqlDbConnecter);
begin
 inherited Create;
 myConnecter:=aConnecter;
end;

function TSqlDbCommander.GetConnecter:TSqlDbConnecter;
begin
 if Assigned(Self)
 then Result:=myConnecter
 else Result:=nil;
end;

function TSqlDbCommander.State:Integer;
begin
 Result:=adStateClosed;
 if Assigned(Self) then
 try
  Result:=Connecter.State;
 except
  on E:Exception do BugReport(E,Self,'State');
 end;
end;

function TSqlDbCommander.Active:Boolean;
begin
 Result:=False;
 if Assigned(Self) then
 try
  Result:=Connecter.Active;
 except
  on E:Exception do BugReport(E,Self,'Active');
 end;
end;

function TSqlDbCommander.Cancel:Boolean;
begin
 Result:=False;
 if Assigned(Self) then
 try
  if Connecter.Api.IsLinked then
  if Connecter.Api.Query.Active then begin
   Connecter.Api.DataSource.DataSet.Cancel;
   Connecter.Api.Query.CancelUpdates;
   Result:=true;
  end;
 except
  on E:Exception do BugReport(E,Self,'Cancel');
 end;
end;

function TSqlDbCommander.GetCommandType:Integer;
begin
 Result:=0;
 // Stub for ADO simulation
 if Assigned(Self) then Result:=adCmdText;
end;

procedure TSqlDbCommander.SetCommandType(t:Integer);
begin
 // Stub for ADO simulation
 if Assigned(Self) then Connecter.Api.ParameterUnused(t);
end;

function TSqlDbCommander.GetCommandText:LongString;
begin
 Result:='';
 if Assigned(Self) then
 try
  if Connecter.Api.IsLinked then begin
   Result:=Connecter.QuerySQLText;
  end;
 except
  on E:Exception do BugReport(E,Self,'GetCommandText');
 end;
end;

procedure TSqlDbCommander.SetCommandText(aCommand:LongString);
begin
 if Assigned(Self) then
 try
  if Connecter.Api.IsLinked then begin
   Connecter.QuerySQLText:=aCommand;
  end;
 except
  on E:Exception do BugReport(E,Self,'SetCommandText');
 end;
end;

////////////////////////////////
// TSqlDbRecorder implementation
////////////////////////////////

constructor TSqlDbRecorder.Create(aConnecter:TSqlDbConnecter);
begin
 inherited Create;
 myConnecter:=aConnecter;
end;

function TSqlDbRecorder.GetConnecter:TSqlDbConnecter;
begin
 if Assigned(Self)
 then Result:=myConnecter
 else Result:=nil;
end;

function TSqlDbRecorder.BOF:Boolean;
begin
 Result:=False;
 if Assigned(Self) then
 try
  if Connecter.Api.IsLinked then
  if Connecter.Api.Query.Active then begin
   Result:=Connecter.Api.DataSource.DataSet.BOF;
  end;
 except
  on E:Exception do BugReport(E,Self,'BOF');
 end;
end;

function TSqlDbRecorder.EOF:Boolean;
begin
 Result:=False;
 if Assigned(Self) then
 try
  if Connecter.Api.IsLinked then
  if Connecter.Api.Query.Active then begin
   Result:=Connecter.Api.DataSource.DataSet.EOF;
  end;
 except
  on E:Exception do BugReport(E,Self,'EOF');
 end;
end;

function TSqlDbRecorder.GetFieldsCount:Integer;
begin
 Result:=0;
 if Assigned(Self) then
 try
  if Connecter.Api.IsLinked then
  if Connecter.Api.Query.Active then begin
   Result:=Connecter.Api.DataSource.DataSet.FieldCount;
  end;
 except
  on E:Exception do BugReport(E,Self,'GetFieldsCount');
 end;
end;

function TSqlDbRecorder.GetFieldsNames(i:Integer):LongString;
begin
 Result:='';
 if Assigned(Self) then
 try
  if Connecter.Api.IsLinked then
  if Connecter.Api.Query.Active then begin
   if InRange(i,0,Connecter.Api.DataSource.DataSet.FieldCount-1)
   then Result:=Connecter.Api.DataSource.DataSet.Fields[i].FieldName;
  end;
 except
  on E:Exception do BugReport(E,Self,'GetFieldsNames');
 end;
end;

function TSqlDbRecorder.GetFieldsTypes(i:Integer):Integer;
begin
 Result:=0;
 if Assigned(Self) then
 try
  if Connecter.Api.IsLinked then
  if Connecter.Api.Query.Active then begin
   if InRange(i,0,Connecter.Api.DataSource.DataSet.FieldCount-1)
   then Result:=Connecter.FieldTypeToCode(Connecter.Api.DataSource.DataSet.Fields[i].DataType);
  end;
 except
  on E:Exception do BugReport(E,Self,'GetFieldsTypes');
 end;
end;

function TSqlDbRecorder.IndexOfField(const id:LongString):Integer;
var Field:TField;
begin
 Result:=-1;
 if (id<>'') then
 if Assigned(Self) then
 try
  Field:=nil;
  if Connecter.Api.IsLinked then
  if Connecter.Api.Query.Active then begin
   Field:=Connecter.Api.DataSource.DataSet.Fields.FindField(id);
   if Assigned(Field) then Result:=Field.Index;
  end;
 except
  on E:Exception do BugReport(E,Self,'IndexOfField');
 end;
end;

function TSqlDbRecorder.MoveFirst:Boolean;
begin
 Result:=False;
 if Assigned(Self) then
 try
  if Connecter.Api.IsLinked then
  if Connecter.Api.Query.Active then begin
   Connecter.Api.DataSource.DataSet.First;
   Result:=true;
  end;
 except
  on E:Exception do BugReport(E,Self,'MoveFirst');
 end;
end;

function TSqlDbRecorder.MoveLast:Boolean;
begin
 Result:=False;
 if Assigned(Self) then
 try
  if Connecter.Api.IsLinked then
  if Connecter.Api.Query.Active then begin
   Connecter.Api.DataSource.DataSet.Last;
   Result:=true;
  end;
 except
  on E:Exception do BugReport(E,Self,'MoveLast');
 end;
end;

function TSqlDbRecorder.MoveNext:Boolean;
begin
 Result:=False;
 if Assigned(Self) then
 try
  if Connecter.Api.IsLinked then
  if Connecter.Api.Query.Active then begin
   Connecter.Api.DataSource.DataSet.Next;
   Result:=true;
  end;
 except
  on E:Exception do BugReport(E,Self,'MoveNext');
 end;
end;

function TSqlDbRecorder.MovePrevious:Boolean;
begin
 Result:=False;
 if Assigned(Self) then
 try
  if Connecter.Api.IsLinked then
  if Connecter.Api.Query.Active then begin
   Connecter.Api.DataSource.DataSet.Prior;
   Result:=true;
  end;
 except
  on E:Exception do BugReport(E,Self,'MovePrevious');
 end;
end;

function TSqlDbRecorder.Close:Boolean;
begin
 Result:=False;
 if Assigned(Self) then
 try
  if Connecter.Api.IsLinked then begin
   if Connecter.Api.Query.Active
   then Connecter.Api.Query.Active:=False;
   Result:=not Connecter.Api.Query.Active;
  end;
 except
  on E:Exception do BugReport(E,Self,'Close');
 end;
end;

function TSqlDbRecorder.Open(opt:Integer):Boolean;
begin
 Result:=False;
 if Assigned(Self) then
 try
  if Connecter.Api.IsLinked then begin
   if not Connecter.Api.Connection.Connected
   then Connecter.Api.Connection.Connected:=True;
   if not Connecter.Api.Query.Active
   then Connecter.Api.Query.Active:=True;
   Result:=Connecter.Api.Query.Active;
   Connecter.Api.ParameterUnused(opt);
  end;
 except
  on E:Exception do BugReport(E,Self,'Open');
 end;
end;

function TSqlDbRecorder.Update:Boolean;
begin
 Result:=False;
 if Assigned(Self) then
 try
  if Connecter.Api.IsLinked then
  if Connecter.Api.Query.Active then begin
   if Connecter.Api.Query.Modified then begin
    Connecter.Api.Query.Post;
    Connecter.Api.Query.ApplyUpdates;
    Result:=true;
   end;
  end;
 except
  on E:Exception do BugReport(E,Self,'Update');
 end;
end;

function TSqlDbRecorder.Cancel:Boolean;
begin
 Result:=False;
 if Assigned(Self) then
 try
  if Connecter.Api.IsLinked then
  if Connecter.Api.Query.Active then begin
   Connecter.Api.DataSource.DataSet.Cancel;
   Connecter.Api.Query.CancelUpdates;
   Result:=true;
  end;
 except
  on E:Exception do BugReport(E,Self,'Cancel');
 end;
end;

function TSqlDbRecorder.CancelUpdate:Boolean;
begin
 Result:=False;
 if Assigned(Self) then
 try
  if Connecter.Api.IsLinked then
  if Connecter.Api.Query.Active then begin
   Connecter.Api.DataSource.DataSet.Cancel;
   Connecter.Api.Query.CancelUpdates;
   Result:=true;
  end;
 except
  on E:Exception do BugReport(E,Self,'CancelUpdate');
 end;
end;

function TSqlDbRecorder.State:Integer;
begin
 Result:=0;
 if Assigned(Self) then
 try
  if Connecter.Api.IsLinked then begin
   Result:=Ord(Connecter.Api.Query.State);
  end;
 except
  on E:Exception do BugReport(E,Self,'State');
 end;
end;

function TSqlDbRecorder.Active:Boolean;
begin
 Result:=False;
 if Assigned(Self) then
 try
  if Connecter.Api.IsLinked then
  if Connecter.Api.Connection.Connected then begin
   Result:=Connecter.Api.Query.Active;
  end;
 except
  on E:Exception do BugReport(E,Self,'Active');
 end;
end;

function TSqlDbRecorder.Requery(opt:Integer=0):Boolean;
begin
 Result:=False;
 if Assigned(Self) then
 try
  if Connecter.Api.IsLinked then
  if Connecter.Api.Connection.Connected then
  if Connecter.Api.Query.SQL.Count>0 then begin
   if Connecter.Api.Query.Active
   then Connecter.Api.Query.Active:=False;
   Connecter.Api.Query.Active:=True;
   Result:=Connecter.Api.Query.Active;
  end;
  Connecter.Api.ParameterUnused(opt);
 except
  on E:Exception do BugReport(E,Self,'Requery');
 end;
end;

function TSqlDbRecorder.Resync(aff,res:Integer):Boolean;
begin
 Result:=False;
 if Assigned(Self) then
 try
  if Connecter.Api.IsLinked then
  if Connecter.Api.Query.Active then begin
   Connecter.Api.DataSource.DataSet.Refresh;
   Result:=True;
  end;
  Connecter.Api.ParameterUnused(aff);
  Connecter.Api.ParameterUnused(res);
 except
  on E:Exception do BugReport(E,Self,'Resync');
 end;
end;

function TSqlDbRecorder.Supports(CursorOptions:Integer):Boolean;
var yes,non:Integer;
begin
 Result:=False;
 if Assigned(Self) then
 try
  yes:=0; non:=0;
  if Connecter.Api.IsLinked then begin
   if HasFlags(CursorOptions,adHoldRecords) then begin
    inc(yes);
   end;
   if HasFlags(CursorOptions,adMovePrevious) then begin
    inc(yes);
   end;
   if HasFlags(CursorOptions,adAddNew) then begin
    inc(yes);
   end;
   if HasFlags(CursorOptions,adDelete) then begin
    inc(yes);
   end;
   if HasFlags(CursorOptions,adUpdate) then begin
    inc(yes);
   end;
   if HasFlags(CursorOptions,adBookmark) then begin
    inc(yes);
   end;
   if HasFlags(CursorOptions,adApproxPosition) then begin
    inc(yes);
   end;
   if HasFlags(CursorOptions,adUpdateBatch) then begin
    inc(yes);
   end;
   if HasFlags(CursorOptions,adResync) then begin
    inc(yes);
   end;
   if HasFlags(CursorOptions,adNotify) then begin
    inc(yes);
   end;
   if HasFlags(CursorOptions,adFind) then begin
    inc(yes);
   end;
   if HasFlags(CursorOptions,adSeek) then begin
    inc(yes);
   end;
   if HasFlags(CursorOptions,adIndex) then begin
    inc(yes);
   end;
   Result:=(yes>0) and (non=0);
  end;
 except
  on E:Exception do BugReport(E,Self,'Supports');
 end;
end;

function TSqlDbRecorder.AddNew(arg:LongString):Boolean;
var Lines:TStringList; i:Integer; sn,sv:LongString; Field:TField;
begin
 Result:=False;
 if Assigned(Self) then
 try
  if Connecter.Api.IsLinked then begin
   if Connecter.Api.Query.Active then begin
    Connecter.Api.DataSource.DataSet.Append;
    Result:=true;
   end;
   arg:=Trim(arg);
   if (arg<>'') then
   if Result then begin
    Lines:=TStringList.Create;
    try
     Lines.Text:=arg;
     for i:=0 to Lines.Count-1 do begin
      if ExtractNameValuePair(Lines[i],sn,sv)>0 then begin
       Field:=Connecter.Api.DataSource.DataSet.FindField(sn);
       if not Assigned(Field) then Continue;
       Field.AsString:=sv;
      end;
     end;
    finally
     Kill(Lines);
    end;
   end;
  end;
 except
  on E:Exception do BugReport(E,Self,'AddNew');
 end;
end;

function TSqlDbRecorder.Delete(aff:Integer):Boolean;
begin
 Result:=False;
 if Assigned(Self) then
 try
  if Connecter.Api.IsLinked then
  if Connecter.Api.Query.Active then begin
   Connecter.Api.DataSource.DataSet.Delete;
   Result:=true;
  end;
  Connecter.Api.ParameterUnused(aff);
 except
  on E:Exception do BugReport(E,Self,'Delete');
 end;
end;

function TSqlDbRecorder.GetString(n:Integer; coldel,rowdel,nullexpr:LongString):LongString;
var i,rows,cols:Integer; line,item:LongString; Lines,Items:TStringList;
var bm:TBookMark;
begin
 Result:='';
 if Assigned(Self) then
 try
  if Connecter.Api.IsLinked then
  if Connecter.Api.Query.Active then begin
   bm:=Connecter.Api.DataSource.DataSet.Bookmark;
   Lines:=TStringList.Create;
   Items:=TStringList.Create;
   try
    rows:=0;
    while (rows<n) and not Connecter.Api.DataSource.DataSet.EOF do begin
     Items.Clear;
     cols:=Connecter.Api.DataSource.DataSet.FieldCount;
     for i:=0 to cols-1 do begin
      if Connecter.Api.DataSource.DataSet.Fields[i].IsNull
      then item:=nullexpr
      else item:=Connecter.Api.DataSource.DataSet.Fields[i].AsString;
      Items.Add(item);
     end;
     line:=Items.Text;
     if EndsText(EOL,line)
     then line:=Copy(line,1,Length(line)-Length(EOL));
     line:=StringReplace(line,EOL,coldel,[rfReplaceAll]);
     Connecter.Api.DataSource.DataSet.Next;
     Lines.Add(line);
     inc(rows);
    end;
    Result:=Lines.Text;
    if (rowdel<>EOL) then Result:=StringReplace(Result,EOL,rowdel,[rfReplaceAll]);
   finally
    Kill(Lines);
    Kill(Items);
   end;
   Connecter.Api.DataSource.DataSet.Bookmark:=bm;
  end;
 except
  on E:Exception do BugReport(E,Self,'GetFieldsAsString');
 end;
end;

function TSqlDbRecorder.GetFieldsAsString(i:Integer):LongString;
begin
 Result:='';
 if Assigned(Self) then
 try
  if Connecter.Api.IsLinked then
  if Connecter.Api.Query.Active then begin
   if InRange(i,0,Connecter.Api.DataSource.DataSet.FieldCount-1)
   then Result:=Connecter.Api.DataSource.DataSet.Fields[i].AsString;
  end;
 except
  on E:Exception do BugReport(E,Self,'GetFieldsAsString');
 end;
end;

procedure TSqlDbRecorder.SetFieldsAsString(i:Integer; const v:LongString);
begin
 if Assigned(Self) then
 try
  if Connecter.Api.IsLinked then
  if Connecter.Api.Query.Active then begin
   if InRange(i,0,Connecter.Api.DataSource.DataSet.FieldCount-1)
   then Connecter.Api.DataSource.DataSet.Fields[i].AsString:=v;
  end;
 except
  on E:Exception do BugReport(E,Self,'SetFieldsAsString');
 end;
end;

function TSqlDbRecorder.GetFieldsAsInteger(i:Integer):Integer;
begin
 Result:=0;
 if Assigned(Self) then
 try
  if Connecter.Api.IsLinked then
  if Connecter.Api.Query.Active then begin
   if InRange(i,0,Connecter.Api.DataSource.DataSet.FieldCount-1)
   then Result:=Connecter.Api.DataSource.DataSet.Fields[i].AsInteger;
  end;
 except
  on E:Exception do BugReport(E,Self,'GetFieldsAsInteger');
 end;
end;

procedure TSqlDbRecorder.SetFieldsAsInteger(i:Integer; v:Integer);
begin
 if Assigned(Self) then
 try
  if Connecter.Api.IsLinked then
  if Connecter.Api.Query.Active then begin
   if InRange(i,0,Connecter.Api.DataSource.DataSet.FieldCount-1)
   then Connecter.Api.DataSource.DataSet.Fields[i].AsInteger:=v;
  end;
 except
  on E:Exception do BugReport(E,Self,'SetFieldsAsInteger');
 end;
end;

function TSqlDbRecorder.GetFieldsAsFloat(i:Integer):Double;
begin
 Result:=0;
 if Assigned(Self) then
 try
  if Connecter.Api.IsLinked then
  if Connecter.Api.Query.Active then begin
   if InRange(i,0,Connecter.Api.DataSource.DataSet.FieldCount-1)
   then Result:=Connecter.Api.DataSource.DataSet.Fields[i].AsFloat;
  end;
 except
  on E:Exception do BugReport(E,Self,'GetFieldsAsFloat');
 end;
end;

procedure TSqlDbRecorder.SetFieldsAsFloat(i:Integer; v:Double);
begin
 if Assigned(Self) then
 try
  if Connecter.Api.IsLinked then
  if Connecter.Api.Query.Active then begin
   if InRange(i,0,Connecter.Api.DataSource.DataSet.FieldCount-1)
   then Connecter.Api.DataSource.DataSet.Fields[i].AsFloat:=v;
  end;
 except
  on E:Exception do BugReport(E,Self,'SetFieldsAsFloat');
 end;
end;

function TSqlDbRecorder.GetSource:LongString;
begin
 Result:='';
 if Assigned(Self) then
 try
  if Connecter.Api.IsLinked then begin
   Result:=Connecter.QuerySQLText;
  end;
 except
  on E:Exception do BugReport(E,Self,'GetSource');
 end;
end;

procedure TSqlDbRecorder.SetSource(aSource:LongString);
begin
 if Assigned(Self) then
 try
  if Connecter.Api.IsLinked then begin
   Connecter.QuerySQLText:=aSource;
  end;
 except
  on E:Exception do BugReport(E,Self,'SetSource');
 end;
end;

function TSqlDbRecorder.GetAbsolutePage:Integer;
begin
 Result:=0;
 // Stub for ADO simulation
end;

procedure TSqlDbRecorder.SetAbsolutePage(p:Integer);
begin
 // Stub for ADO simulation
 if Assigned(Self) then begin
  Connecter.Api.ParameterUnused(p);
 end;
end;

function TSqlDbRecorder.GetAbsolutePosition:Integer;
begin
 Result:=0;
 // Stub for ADO simulation
end;

procedure TSqlDbRecorder.SetAbsolutePosition(p:Integer);
begin
 // Stub for ADO simulation
 if Assigned(Self) then begin
  Connecter.Api.ParameterUnused(p);
 end;
end;

function TSqlDbRecorder.GetBookmark:Integer;
begin
 Result:=0;
 // Stub for ADO simulation
end;

procedure TSqlDbRecorder.SetBookmark(p:Integer);
begin
 // Stub for ADO simulation
 if Assigned(Self) then begin
  Connecter.Api.ParameterUnused(p);
 end;
end;

function TSqlDbRecorder.GetCacheSize:Integer;
begin
 Result:=0;
 // Stub for ADO simulation
end;

procedure TSqlDbRecorder.SetCacheSize(p:Integer);
begin
 // Stub for ADO simulation
 if Assigned(Self) then begin
  Connecter.Api.ParameterUnused(p);
 end;
end;

function TSqlDbRecorder.GetCursorType:Integer;
begin
 Result:=0;
 // Stub for ADO simulation
end;

procedure TSqlDbRecorder.SetCursorType(p:Integer);
begin
 // Stub for ADO simulation
 if Assigned(Self) then begin
  Connecter.Api.ParameterUnused(p);
 end;
end;

function TSqlDbRecorder.GetEditMode:Integer;
begin
 Result:=0;
 // Stub for ADO simulation
end;

procedure TSqlDbRecorder.SetEditMode(p:Integer);
begin
 // Stub for ADO simulation
 if Assigned(Self) then begin
  Connecter.Api.ParameterUnused(p);
 end;
end;

function TSqlDbRecorder.GetFilter:LongString;
begin
 Result:='';
 // Stub for ADO simulation
end;

procedure TSqlDbRecorder.SetFilter(v:LongString);
begin
 // Stub for ADO simulation
 if Assigned(Self) then begin
  Connecter.Api.ParameterUnused(v);
 end;
end;

function TSqlDbRecorder.GetIndex:LongString;
begin
 Result:='';
 // Stub for ADO simulation
end;

procedure TSqlDbRecorder.SetIndex(v:LongString);
begin
 // Stub for ADO simulation
 if Assigned(Self) then begin
  Connecter.Api.ParameterUnused(v);
 end;
end;

function TSqlDbRecorder.GetLockType:Integer;
begin
 Result:=0;
 // Stub for ADO simulation
end;

procedure TSqlDbRecorder.SetLockType(p:Integer);
begin
 // Stub for ADO simulation
 if Assigned(Self) then begin
  Connecter.Api.ParameterUnused(p);
 end;
end;

function TSqlDbRecorder.GetMarshalOptions:Integer;
begin
 Result:=0;
 // Stub for ADO simulation
end;

procedure TSqlDbRecorder.SetMarshalOptions(p:Integer);
begin
 // Stub for ADO simulation
 if Assigned(Self) then begin
  Connecter.Api.ParameterUnused(p);
 end;
end;

function TSqlDbRecorder.GetMaxRecords:Integer;
begin
 Result:=0;
 // Stub for ADO simulation
end;

procedure TSqlDbRecorder.SetMaxRecords(p:Integer);
begin
 // Stub for ADO simulation
 if Assigned(Self) then begin
  Connecter.Api.ParameterUnused(p);
 end;
end;

function TSqlDbRecorder.GetPageCount:Integer;
begin
 Result:=0;
 // Stub for ADO simulation
end;

function TSqlDbRecorder.GetPageSize:Integer;
begin
 Result:=0;
 // Stub for ADO simulation
end;

procedure TSqlDbRecorder.SetPageSize(p:Integer);
begin
 // Stub for ADO simulation
 if Assigned(Self) then begin
  Connecter.Api.ParameterUnused(p);
 end;
end;

function TSqlDbRecorder.GetRecordCount:Integer;
begin
 Result:=0;
 // Stub for ADO simulation
end;

function TSqlDbRecorder.GetSort:LongString;
begin
 Result:='';
 // Stub for ADO simulation
end;

procedure TSqlDbRecorder.SetSort(v:LongString);
begin
 // Stub for ADO simulation
 if Assigned(Self) then begin
  Connecter.Api.ParameterUnused(v);
 end;
end;

function TSqlDbRecorder.GetStatus:Integer;
begin
 Result:=0;
 // Stub for ADO simulation
end;

function TSqlDbRecorder.GetStayInSync:Boolean;
begin
 Result:=False;
 // Stub for ADO simulation
end;

procedure TSqlDbRecorder.SetStayInSync(v:Boolean);
begin
 // Stub for ADO simulation
 if Assigned(Self) then begin
  Connecter.Api.ParameterUnused(v);
 end;
end;

function TSqlDbRecorder.GetCursorLocation:Integer;
begin
 Result:=0;
 // Stub for ADO simulation
end;

procedure TSqlDbRecorder.SetCursorLocation(p:Integer);
begin
 // Stub for ADO simulation
 if Assigned(Self) then begin
  Connecter.Api.ParameterUnused(p);
 end;
end;

function TSqlDbRecorder.Save(Destination:LongString; PersistFormat:Integer):Boolean;
var df:TDataPacketFormat; ext:LongString;
const tabExts='.txt;.tab;';
const csvExts='.csv;';
begin
 Result:=False;
 if Assigned(Self) then
 try
  if IsNonEmptyStr(Destination) then
  if Connecter.Api.IsLinked then begin
   if InRange(PersistFormat,Ord(Low(df)),Ord(High(df))) then begin
    df:=TDataPacketFormat(PersistFormat);
    if (df in [dfAny]) then begin
     ext:=ExtractFileExt(Destination);
     if WordIndex(ext,csvExts,ScanSpaces)>0 then begin
      Result:=SaveAsText(Destination,',',EOL,'',QuoteMark);
      Exit;
     end;
     if WordIndex(ext,tabExts,ScanSpaces)>0 then begin
      Result:=SaveAsText(Destination,ASCII_Tab,EOL,'',' ');
      Exit;
     end;
    end;
    Connecter.Api.Query.SaveToFile(Destination,df);
    Result:=True;
   end;
  end;
 except
  on E:Exception do BugReport(E,Self,'Save');
 end;
end;

function TSqlDbRecorder.SaveAsText(Dest,coldel,rowdel,nullexpr:LongString;Quote:Char):Boolean;
var i,rows,cols:Integer; line,item,Buff:LongString; Lines,Items:TStringList;
var bm:TBookMark; T:Text; NeedQuotes:Boolean;
const PageLines=1024;
begin
 Result:=False;
 if Assigned(Self) then
 try
  if IsNonEmptyStr(Dest) then
  if Connecter.Api.Query.Active then begin
   bm:=Connecter.Api.DataSource.DataSet.Bookmark;
   Connecter.Api.DataSource.DataSet.First;
   Assign(T,Dest); Rewrite(T);
   Buff:=StringBuffer(32*1024);
   SetTextBuf(T,PChar(Buff)^,Length(Buff));
   Lines:=TStringList.Create;
   Items:=TStringList.Create;
   try
    rows:=0;
    while not Connecter.Api.DataSource.DataSet.EOF do begin
     Items.Clear;
     cols:=Connecter.Api.DataSource.DataSet.FieldCount;
     for i:=0 to cols-1 do begin
      if Connecter.Api.DataSource.DataSet.Fields[i].IsNull
      then item:=nullexpr
      else item:=Connecter.Api.DataSource.DataSet.Fields[i].AsString;
      if Quote in [QuoteMark,Apostrophe] then begin
       NeedQuotes:=HasChars(item,ScanSpaces);
       if NeedQuotes then item:=AnsiQuotedStr(item,Quote);
      end;
      Items.Add(item);
     end;
     line:=Items.Text;
     if EndsText(EOL,line)
     then line:=Copy(line,1,Length(line)-Length(EOL));
     line:=StringReplace(line,EOL,coldel,[rfReplaceAll]);
     Connecter.Api.DataSource.DataSet.Next;
     Lines.Add(line);
     inc(rows);
     if Connecter.Api.DataSource.DataSet.EOF or (Lines.Count>=PageLines) then begin
      line:=lines.Text;
      if (rowdel<>EOL) then line:=StringReplace(line,EOL,rowdel,[rfReplaceAll]);
      write(T,line);
      Lines.Clear;
     end;
    end;
    Result:=(IOResult=0);
   finally
    Kill(Lines);
    Kill(Items);
    SmartFileClose(T);
    Buff:='';
   end;
   Connecter.Api.DataSource.DataSet.Bookmark:=bm;
  end;
 except
  on E:Exception do BugReport(E,Self,'SaveAsText');
 end;
end;

////////////////////////////////////
// AllSqlDbConnecters implementation
////////////////////////////////////

const
 TheSqlDbConnecters:TObjectStorage=nil;

function AllSqlDbConnecters:TObjectStorage;
begin
 if (TheSqlDbConnecters=nil) then begin
  TheSqlDbConnecters:=NewObjectStorage(false);
  TheSqlDbConnecters.Master:=@TheSqlDbConnecters;
  TheSqlDbConnecters.OwnsObjects:=false;
 end;
 Result:=TheSqlDbConnecters;
end;

///////////////////////////////////////
// Unit initialization and finalization
///////////////////////////////////////

procedure Init_crw_dblaz;
begin
 dlc_SqlDbBug;
 dlc_SqlDbLog;
 AllSqlDbConnecters.Ok;
 TSqlDbApi.DefVerbose:=True;
 TSqlDbApi.RaiseOnNilReference:=True;
 TSqlDbApi.RaiseOnResetProvider:=True;
 TSqlDbApi.RaiseOnResetConStrng:=True;
 TSqlDbConnecter.DefPropertiesWanted:='Connection,Transaction,Query,DataSource';
 TSqlDbConnecter.SelectableStatementTypes:=[stSelect];
 TSqlDbConnecter.DefLogEvents:=[Low(TDBEventType)..High(TDBEventType)];
end;

procedure Free_crw_dblaz;
begin
 Kill(TObject(TheSqlDbConnecters));
end;

initialization

 Init_crw_dblaz;

finalization

 Free_crw_dblaz;

end.

//////////////
// END OF FILE
//////////////

